/*
 * Copyright (C) 1998, 1999, 2000, 2001
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 *
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.
 */


/* -------------------------------------------------------------------
 * project ....: LTI-Lib: Image Processing and Computer Vision Library
 * file .......: ltiColorContrastGradient.cpp
 * authors ....: Pablo Alvarado
 * organization: LTI, RWTH Aachen
 * creation ...: 1.6.2003
 * revisions ..: $Id: ltiColorContrastGradient.cpp,v 1.12 2003/11/14 14:15:19 alvarado Exp $
 */

#include "ltiColorContrastGradient.h"
#include "ltiSplitImageToRGB.h"
#include "ltiArctanLUT.h"
#include "ltiGaussKernels.h"
#include "ltiGradientKernels.h"
#include "ltiConvolution.h"
#include "ltiOgdFilter.h"
#include  <limits>

namespace lti {
  // --------------------------------------------------
  // colorContrastGradient::parameters
  // --------------------------------------------------

  // default constructor
  colorContrastGradient::parameters::parameters()
    : transform::parameters() {
    format = eOutputFormat(Polar);
    optimalGradientKernelSize = 0;
    useOGDKernel = false;
    ogdVariance = -1.0f;
  }

  // copy constructor
  colorContrastGradient::parameters::parameters(const parameters& other)
    : transform::parameters() {
    copy(other);
  }

  // destructor
  colorContrastGradient::parameters::~parameters() {
  }

  // get type name
  const char* colorContrastGradient::parameters::getTypeName() const {
    return "colorContrastGradient::parameters";
  }

  // copy member

  colorContrastGradient::parameters&
    colorContrastGradient::parameters::copy(const parameters& other) {
# ifndef _LTI_MSC_VER
    // MS Visual C++ 6 is not able to compile this...
    transform::parameters::copy(other);
# else
    // ...so we have to use this workaround.
    // Conditional on that, copy may not be virtual.
    transform::parameters& (transform::parameters::* p_copy)
      (const transform::parameters&) =
      transform::parameters::copy;
    (this->*p_copy)(other);
# endif

    format = other.format;
    optimalGradientKernelSize = other.optimalGradientKernelSize;
    useOGDKernel = other.useOGDKernel;
    ogdVariance=other.ogdVariance;
    return *this;
  }

  // alias for copy member
  colorContrastGradient::parameters&
    colorContrastGradient::parameters::operator=(const parameters& other) {
    return copy(other);
  }

  // clone member
  functor::parameters* colorContrastGradient::parameters::clone() const {
    return new parameters(*this);
  }

  /*
   * write the parameters in the given ioHandler
   * @param handler the ioHandler to be used
   * @param complete if true (the default) the enclosing begin/end will
   *        be also written, otherwise only the data block will be written.
   * @return true if write was succeful
   */
# ifndef _LTI_MSC_VER
  bool colorContrastGradient::parameters::write(ioHandler& handler,
                                         const bool& complete) const
# else
  bool colorContrastGradient::parameters::writeMS(ioHandler& handler,
                                           const bool& complete) const
# endif
  {
    bool b = true;
    if (complete) {
      b = handler.writeBegin();
    }

    if (b) {
      switch(format){
        case Polar:
          lti::write(handler,"format","Polar");
          break;
        case Cartesic:
          lti::write(handler,"format","Cartesic");
          break;
        default:
          lti::write(handler,"format","Polar");          
      }

      lti::write(handler,"optimalGradientKernelSize",
                 optimalGradientKernelSize);

      lti::write(handler,"useOGDKernel",useOGDKernel);
      lti::write(handler,"ogdVariance",ogdVariance);
    }

# ifndef _LTI_MSC_VER
    // This is the standard C++ code, which MS Visual C++ 6 is not able to
    // compile...
    b = b && transform::parameters::write(handler,false);
# else
    bool (transform::parameters::* p_writeMS)(ioHandler&,const bool&) const =
      transform::parameters::writeMS;
    b = b && (this->*p_writeMS)(handler,false);
# endif

    if (complete) {
      b = b && handler.writeEnd();
    }

    return b;
  }

# ifdef _LTI_MSC_VER
  bool colorContrastGradient::parameters::write(ioHandler& handler,
                                         const bool& complete) const {
    // ...we need this workaround to cope with another really awful MSVC bug.
    return writeMS(handler,complete);
  }
# endif

  /*
   * read the parameters from the given ioHandler
   * @param handler the ioHandler to be used
   * @param complete if true (the default) the enclosing begin/end will
   *        be also read, otherwise only the data block will be read.
   * @return true if write was succeful
   */
# ifndef _LTI_MSC_VER
  bool colorContrastGradient::parameters::read(ioHandler& handler,
                                        const bool& complete)
# else
  bool colorContrastGradient::parameters::readMS(ioHandler& handler,
                                          const bool& complete)
# endif
  {
    bool b = true;
    if (complete) {
      b = handler.readBegin();
    }

    if (b) {
      std::string str;
      lti::read(handler,"format",str);
      if (str == "Cartesic") {
        format = Cartesic;
      } else {
        format = Polar;
      }
      lti::read(handler,"optimalGradientKernelSize",optimalGradientKernelSize);
      lti::read(handler,"useOGDKernel",useOGDKernel);
      lti::read(handler,"ogdVariance",ogdVariance);
    }

# ifndef _LTI_MSC_VER
    // This is the standard C++ code, which MS Visual C++ 6 is not able to
    // compile...
    b = b && transform::parameters::read(handler,false);
# else
    bool (transform::parameters::* p_readMS)(ioHandler&,const bool&) =
      transform::parameters::readMS;
    b = b && (this->*p_readMS)(handler,false);
# endif

    if (complete) {
      b = b && handler.readEnd();
    }

    return b;
  }

# ifdef _LTI_MSC_VER
  bool colorContrastGradient::parameters::read(ioHandler& handler,
                                        const bool& complete) {
    // ...we need this workaround to cope with another really awful MSVC
    // bug.
    return readMS(handler,complete);
  }
# endif

  // --------------------------------------------------
  // colorContrastGradient
  // --------------------------------------------------

  // default constructor
  colorContrastGradient::colorContrastGradient() {
    parameters defaultParameters;
    // set the default parameters
    setParameters(defaultParameters);
  }

  // default constructor
  colorContrastGradient::colorContrastGradient(const parameters::eOutputFormat& f,
                                               const int gradKernelSize)
    : transform() {

    // create an instance of the parameters with the default values
    parameters defaultParameters;
    defaultParameters.format=f;
    defaultParameters.optimalGradientKernelSize = gradKernelSize;

    // set the default parameters
    setParameters(defaultParameters);
  }

  // default constructor
  colorContrastGradient::colorContrastGradient(const parameters& par)
    : transform() {
    // set the given parameters
    setParameters(par);
  }


  // copy constructor
  colorContrastGradient::colorContrastGradient(const colorContrastGradient& other)
    : transform() {
    copy(other);
  }

  // destructor
  colorContrastGradient::~colorContrastGradient() {
  }

  // returns the name of this type
  const char* colorContrastGradient::getTypeName() const {
    return "colorContrastGradient";
  }

  // which apply is faster
  functor::eFasterApply colorContrastGradient::theFaster() const {
    return OnCopy;
  }

  // copy member
  colorContrastGradient&
  colorContrastGradient::copy(const colorContrastGradient& other) {
    transform::copy(other);

    return (*this);
  }

  // alias for copy member
  colorContrastGradient&
  colorContrastGradient::operator=(const colorContrastGradient& other) {
    return (copy(other));
  }


  // clone member
  functor* colorContrastGradient::clone() const {
    return new colorContrastGradient(*this);
  }

  // return parameters
  const colorContrastGradient::parameters&
  colorContrastGradient::getParameters() const {
    const parameters* par =
      dynamic_cast<const parameters*>(&functor::getParameters());
    if(isNull(par)) {
      throw invalidParametersException(getTypeName());
    }
    return *par;
  }

  // -------------------------------------------------------------------
  // The apply-methods!
  // -------------------------------------------------------------------
  
  // On copy apply for type image!
  bool colorContrastGradient::apply(const image& src,
                                    channel& xOrMag,
                                    channel& yOrArg) const {
    splitImageToRGB splitter;
    channel c1,c2,c3;

    splitter.apply(src,c1,c2,c3);
    float maxmag;
    return apply(c1,c2,c3,xOrMag,yOrArg,maxmag);
  };

  bool colorContrastGradient::apply(const channel& c1,
                                    const channel& c2,
                                    const channel& c3,
                                    channel& xOrMag,
                                    channel& yOrArg) const {
    float maxmag;
    return apply(c1,c2,c3,xOrMag,yOrArg,maxmag);
  }

  bool colorContrastGradient::apply(const channel& c1,
                                    const channel& c2,
                                    const channel& c3,
                                    channel& xOrMag,
                                    channel& yOrArg,
                                    float& maxmag) const {

    if (getParameters().format == parameters::Polar) {
      return computeGradientPolar(c1,c2,c3,xOrMag,yOrArg,maxmag);
    } else {
      return computeGradientCart(c1,c2,c3,xOrMag,yOrArg,maxmag);      
    }
  }


  bool colorContrastGradient::apply(const channel& c1,
                                    const channel& c2,
                                    channel& xOrMag,
                                    channel& yOrArg,
                                    float& maxmag) const {

    if (getParameters().format == parameters::Polar) {
      return computeGradientPolar(c1,c2,xOrMag,yOrArg,maxmag);
    } else {
      return computeGradientCart(c1,c2,xOrMag,yOrArg,maxmag);      
    }
  }


  bool colorContrastGradient::apply(const channel& c1,
                                    const channel& c2,
                                    const channel& c3,
                                    channel& lambdaMin,
                                    channel& lambdaMax,
                                    channel& orientation,
                                    float& maxmag) const {

    return computeGradientPolar(c1,c2,c3,lambdaMin,lambdaMax,orientation,
                                maxmag);
  }
    
  bool colorContrastGradient::apply(const channel& src,
                                    channel& xOrMag,
                                    channel& yOrArg) const {
    
    if (computeGradientCart(src,xOrMag,yOrArg)) {
      if (getParameters().format == parameters::Polar) {
        int y,x;
        float cx,cy;
        for (y=0;y<src.rows();++y) {
          for (x=0;x<src.columns();++x) {
            cx=xOrMag.at(y,x);
            cy=yOrArg.at(y,x);
            xOrMag.at(y,x) = sqrt(cx*cx + cy*cy);
            yOrArg.at(y,x) = atan2(cy,cx);
          }
        }
      }
    }

    return true;
  }

  /*
   * differentiate the image in X and Y directions
   * @param c1 first channel
   * @param c2 second channel
   * @param c3 third channel
   * @param gradAbs the absolute values of the gradient
   * @param orientation the orientation map of the gradient image
   * @return true if successful, false otherwise.
   */
  bool colorContrastGradient::computeGradientPolar(const channel& c1,
                                                   const channel& c2,
                                                   const channel& c3,
                                                   channel& gradAbs,
                                                   channel& orientation,
                                                   float& maxVal) const {
    
    // compute the gradients
    channel dx1,dx2,dx3,dy1,dy2,dy3;

    // compute the partial derivatives of each channel respect to x and y
    computeGradientCart(c1,dx1,dy1);
    computeGradientCart(c2,dx2,dy2);
    computeGradientCart(c3,dx3,dy3);

    // resize the result
    gradAbs.resize(c1.size(),float(),false,false);
    orientation.resize(c1.size(),float(),false,false);

    // Determine the direction and amplitude of highest contrast.
    int x,y;
    float E,F,G,EpG,EmG,a;
    trgbPixel<float> dfdx(0,0,0),dfdy(0,0,0);
    maxVal=0.0f;

    for (y=0;y<c1.rows();++y) {
      for (x=0;x<c1.columns();++x) {
        // get both derivative vectors
        dfdx.set(dx1.at(y,x),dx2.at(y,x),dx3.at(y,x));
        dfdy.set(dy1.at(y,x),dy2.at(y,x),dy3.at(y,x));

        // components of the tensor
        E=dfdx.dot(dfdx);
        F=dfdx.dot(dfdy);
        G=dfdy.dot(dfdy);

        // the vector of maximal contrast is given by the eigenvalues of
        // the tensor matrix lambda = (E+G) +/- sqrt((E-G)^2+4F^2)/2;
        EpG=E+G;
        if (EpG<=std::numeric_limits<float>::epsilon()) {
          gradAbs.at(y,x)=0.0f;
          orientation.at(y,x)=0.0f;
        } else {
          EmG=E-G;
//           maxVal=max((gradAbs.at(y,x)=EpG+sqrt(EmG*EmG+4.0f*F*F)),maxVal);
          // best contrast obtained with lambda_max - lambda_min
          maxVal=max((gradAbs.at(y,x)=pow(EmG*EmG+4.0f*F*F,0.25f)),maxVal);
          a = 0.5f*atan2(2.0f*F,EmG);
          orientation.at(y,x)=a;
        }
      }
    }

    if (getParameters().optimalGradientKernelSize == 2) {
      // we still need to fix the angle, which is shifted 45 degrees:
      orientation.add(static_cast<float>(Pi/4.0));
    }
    
    return true;
  }

  /*
   * differentiate the image in X and Y directions
   * @param c1 first channel
   * @param c2 second channel
   * @param gradAbs the absolute values of the gradient
   * @param orientation the orientation map of the gradient image
   * @return true if successful, false otherwise.
   */
  bool colorContrastGradient::computeGradientPolar(const channel& c1,
                                                   const channel& c2,
                                                   channel& gradAbs,
                                                   channel& orientation,
                                                   float& maxVal) const {
    
    // compute the gradients
    channel dx1,dx2,dy1,dy2;

    // compute the partial derivatives of each channel respect to x and y
    computeGradientCart(c1,dx1,dy1);
    computeGradientCart(c2,dx2,dy2);

    // resize the result
    gradAbs.resize(c1.size(),float(),false,false);
    orientation.resize(c1.size(),float(),false,false);

    // Determine the direction and amplitude of highest contrast.
    int x,y;
    float E,F,G,EpG,EmG,a;
    tpoint<float> dfdx(0,0),dfdy(0,0);
    maxVal=0.0f;

    for (y=0;y<c1.rows();++y) {
      for (x=0;x<c1.columns();++x) {
        // get both derivative vectors
        dfdx.set(dx1.at(y,x),dx2.at(y,x));
        dfdy.set(dy1.at(y,x),dy2.at(y,x));

        // components of the tensor
        E=dfdx.dot(dfdx);
        F=dfdx.dot(dfdy);
        G=dfdy.dot(dfdy);

        // the vector of maximal contrast is given by the eigenvalues of
        // the tensor matrix lambda = (E+G) +/- sqrt((E-G)^2+4F^2)/2;
        EpG=E+G;
        if (EpG<=std::numeric_limits<float>::epsilon()) {
          gradAbs.at(y,x)=0.0f;
          orientation.at(y,x)=0.0f;
        } else {
          EmG=E-G;
//           maxVal=max((gradAbs.at(y,x)=EpG+sqrt(EmG*EmG+4.0f*F*F)),maxVal);
          // best contrast obtained with lambda_max - lambda_min
          maxVal=max((gradAbs.at(y,x)=pow(EmG*EmG+4.0f*F*F,0.25f)),maxVal);
          a = 0.5f*atan2(2.0f*F,EmG);
          orientation.at(y,x)=a;
        }
      }
    }

    if (getParameters().optimalGradientKernelSize == 2) {
      // we still need to fix the angle, which is shifted 45 degrees:
      orientation.add(static_cast<float>(Pi/4.0));
    }
    
    return true;
  }


  /*
   * differentiate the image in X and Y directions
   *
   * The lambdaMax-lambdaMin corresponds usually to the highest contrast.
   *
   * @param c1 first channel
   * @param c2 second channel
   * @param c3 third channel
   * @param lambdaMin minimum eigenvalue
   * @param lambdaMax maximum eigenvalue
   * @param orientation the orientation map of the gradient image
   * @return true if successful, false otherwise.
   */
  bool colorContrastGradient::computeGradientPolar(const channel& c1,
                                                   const channel& c2,
                                                   const channel& c3,
                                                   channel& lambdaMin,
                                                   channel& lambdaMax,
                                                   channel& orientation,
                                                   float& maxVal) const {
    
    // compute the gradients
    channel dx1,dx2,dx3, dy1,dy2,dy3;

    // compute the partial derivatives of each channel respect to x and y
    computeGradientCart(c1,dx1,dy1);
    computeGradientCart(c2,dx2,dy2);
    computeGradientCart(c3,dx3,dy3);

    // resize the result
    lambdaMin.resize(c1.size(),float(),false,false);
    lambdaMax.resize(c1.size(),float(),false,false);
    orientation.resize(c1.size(),float(),false,false);

    // Determine the direction and amplitude of highest contrast.
    int x,y;
    float E,F,G,EpG,EmG,a;
    trgbPixel<float> dfdx(0,0,0),dfdy(0,0,0);
    maxVal=0.0f;

    for (y=0;y<c1.rows();++y) {
      for (x=0;x<c1.columns();++x) {
        // get both derivative vectors
        dfdx.set(dx1.at(y,x),dx2.at(y,x),dx3.at(y,x));
        dfdy.set(dy1.at(y,x),dy2.at(y,x),dy3.at(y,x));

        // components of the tensor
        E=dfdx.dot(dfdx);
        F=dfdx.dot(dfdy);
        G=dfdy.dot(dfdy);

        // the vector of maximal contrast is given by the eigenvalues of
        // the tensor matrix lambda = (E+G) +/- sqrt((E-G)^2+4F^2)/2;
        EpG=E+G;
        if (EpG <= std::numeric_limits<float>::epsilon()) {
          lambdaMin.at(y,x)=lambdaMax.at(y,x)=orientation.at(y,x)=0.0f;
        } else {
          EmG=E-G;
          a = sqrt(EmG*EmG+4.0f*F*F);
          lambdaMin.at(y,x) = (EpG - a)*0.5f;
          lambdaMax.at(y,x) = (EpG + a)*0.5f;

          // best contrast obtained with lambda_max - lambda_min
          maxVal=max(a,maxVal);
          a = 0.5f*atan2(2.0f*F,EmG);
          orientation.at(y,x)=a;
        }
      }
    }

    if (getParameters().optimalGradientKernelSize == 2) {
      // we still need to fix the angle, which is shifted 45 degrees:
      orientation.add(static_cast<float>(Pi/4.0));
    }
    
    return true;
  }


  bool colorContrastGradient::computeGradientCart(const channel& c1,
                                                  const channel& c2,
                                                  const channel& c3,
                                                  channel& xchnl,
                                                  channel& ychnl,
                                                  float& maxVal) const {
    
    // compute the gradients
    channel dx1,dx2,dx3,dy1,dy2,dy3;

    // compute the partial derivatives of each channel respect to x and y
    computeGradientCart(c1,dx1,dy1);
    computeGradientCart(c2,dx2,dy2);
    computeGradientCart(c3,dx3,dy3);

    // resize the result
    xchnl.resize(c1.size(),float(),false,false);
    ychnl.resize(c1.size(),float(),false,false);

    // Determine the direction and amplitude of highest contrast.

    int x,y;
    float E,F,G,EpG,EmG,mag2;
    trgbPixel<float> dfdx(0,0,0),dfdy(0,0,0);
    maxVal=0.0f;

    for (y=0;y<c1.rows();++y) {
      for (x=0;x<c1.columns();++x) {
        // get both derivative vectors
        dfdx.set(dx1.at(y,x),dx2.at(y,x),dx3.at(y,x));
        dfdy.set(dy1.at(y,x),dy2.at(y,x),dy3.at(y,x));

        // components of the tensor
        E=dfdx.dot(dfdx);
        F=dfdx.dot(dfdy);
        G=dfdy.dot(dfdy);

        // the vector of maximal contrast is given by the eigenvalues of
        // the tensor matrix lambda = (E+G) +/- sqrt((E-G)^2+4F^2)/2;
        EpG=E+G;
        if (EpG<=std::numeric_limits<float>::epsilon()) {
          xchnl.at(y,x)=0.0f;
          ychnl.at(y,x)=0.0f;
        } else {
          EmG=E-G;
          mag2 = sqrt(EmG*EmG+4.0f*F*F);
          
          // best contrast obtained with lambda_max - lambda_min
          xchnl.at(y,x)=sqrt((mag2 + EmG)/2.0f);
          ychnl.at(y,x)=sqrt((mag2 - EmG)/2.0f);
          
          maxVal=max(mag2,maxVal);
        }
      }
    }
    
    maxVal=sqrt(maxVal);

    return true;
  }

  // cartesian bi-spectral
  
  bool colorContrastGradient::computeGradientCart(const channel& c1,
                                                  const channel& c2,
                                                  channel& xchnl,
                                                  channel& ychnl,
                                                  float& maxVal) const {
    
    // compute the gradients
    channel dx1,dx2,dy1,dy2;

    // compute the partial derivatives of each channel respect to x and y
    computeGradientCart(c1,dx1,dy1);
    computeGradientCart(c2,dx2,dy2);

    // resize the result
    xchnl.resize(c1.size(),float(),false,false);
    ychnl.resize(c1.size(),float(),false,false);

    // Determine the direction and amplitude of highest contrast.

    int x,y;
    float E,F,G,EpG,EmG,mag2;
    tpoint<float> dfdx(0,0),dfdy(0,0);
    maxVal=0.0f;

    for (y=0;y<c1.rows();++y) {
      for (x=0;x<c1.columns();++x) {
        // get both derivative vectors
        dfdx.set(dx1.at(y,x),dx2.at(y,x));
        dfdy.set(dy1.at(y,x),dy2.at(y,x));

        // components of the tensor
        E=dfdx.dot(dfdx);
        F=dfdx.dot(dfdy);
        G=dfdy.dot(dfdy);

        // the vector of maximal contrast is given by the eigenvalues of
        // the tensor matrix lambda = (E+G) +/- sqrt((E-G)^2+4F^2)/2;
        EpG=E+G;
        if (EpG<=std::numeric_limits<float>::epsilon()) {
          xchnl.at(y,x)=0.0f;
          ychnl.at(y,x)=0.0f;
        } else {
          EmG=E-G;
          mag2 = sqrt(EmG*EmG+4.0f*F*F);
          
          // best contrast obtained with lambda_max - lambda_min
          xchnl.at(y,x)=sqrt((mag2 + EmG)/2.0f);
          ychnl.at(y,x)=sqrt((mag2 - EmG)/2.0f);
          
          maxVal=max(mag2,maxVal);
        }
      }
    }
    
    maxVal=sqrt(maxVal);

    return true;
  }


   bool colorContrastGradient::computeGradientCart(const channel& src,
                                                   channel& dx,
                                                   channel& dy) const {
     
     const parameters& par = getParameters();
    
     bool result=false;

     if (par.useOGDKernel) {
       ogdFilter::parameters ogdPar;
       ogdPar.size = par.optimalGradientKernelSize;
       ogdPar.variance = par.ogdVariance;
       ogdPar.boundaryType = Constant;
       ogdFilter ogd(ogdPar);

       ogd.generateBasisOgd1(src,dx,dy);

       return true;
       
     } else {
       dx.resize(src.size(),float(),false,false);
       dy.resize(src.size(),float(),false,false);

       //XY-Differentiate Image and get gradient in polar coordinates:
       switch(par.optimalGradientKernelSize){
         case(2):
           //inter-pixel gradient
           result=xyDifferentiateImageCartInterPixel(src,dx,dy);
           // please note that dx and dy are not the usual x and y, but rotated
           // 45 degrees
           break;
         case(3):
         case(4):
         case(5):
           //the optimal gradient kernels
           result=xyDifferentiateImageCartOpt(src,dx,dy);
           break;
         default:
           result=xyDifferentiateImageCart(src,dx,dy);
       }
     }
     
     return result;
  }

  bool
  colorContrastGradient::xyDifferentiateImageCartOpt(const channel& src,
                                                     channel& dx,
                                                     channel& dy) const {

#   ifdef _LTI_DEBUG
    static viewer viewGX("GX"),viewGY("GY");
#   endif

    const parameters& par = getParameters();

    gradientKernelX<channel::value_type> gx(par.optimalGradientKernelSize);
    gradientKernelY<channel::value_type> gy(par.optimalGradientKernelSize);

    convolution grad;
    convolution::parameters gradpar;

    gradpar.boundaryType = lti::Constant;

    // x direction
    gradpar.setKernel(gx);
    grad.setParameters(gradpar);
    grad.apply(src,dx);

    // y direction
    grad.setKernel(gy);
    grad.apply(src,dy);

#   ifdef _LTI_DEBUG
    viewGX.show(dx);
    viewGY.show(dy);
#   endif      

    return true;
  }
  
  bool
  colorContrastGradient::xyDifferentiateImageCart(const channel& src, //in:
                                                  channel& dx,    //out: x
                                                  channel& dy     //out: y
                                                  ) const {

    if (src.columns() < 3) {
      setStatusString("width less than 3");
      return false;
    }
    if (src.rows() < 3) {
      setStatusString("height less than 3");
      return false;
    }

    const int width  = src.columns();
    const int height = src.rows();

    float* fpDx = &dx.at(0,0);
    float* fpDy = &dy.at(0,0);

    const float* fpSrc = &src.at(0,0);
    const float* rowy;
    const float* colx;

    float* pidx;
    float* pidy;

    const int w1 = width-1;
    const int w2 = width-2;
    const int last = (height-1)*width; // index of begin of last row
    const int lastRow = -w1;           // offset from actual column pointer to
                                       // last row
    const int nextRow = width+1;       // offset from actual column pointer to
                                       // next row

    // top-left corner
    fpDx[0]=(fpSrc[1]-fpSrc[0]);
    fpDy[0]=(fpSrc[width]-fpSrc[0]);

    // top
    pidx = &fpDx[1];
    pidy = &fpDy[1];

    for (colx=&fpSrc[0],rowy=&fpSrc[w1];
         colx<rowy;
         ++colx,++pidx,++pidy) {
      *pidx=(colx[2] - *colx);
      *pidy=(colx[nextRow] - colx[1]);
    }

    // top-right corner
    fpDx[w1]=(fpSrc[w1]-fpSrc[w2]);
    fpDy[w1]=(fpSrc[width+w1]-fpSrc[w1]);

    // main loop (begin at coordinates (1,0)
    pidx = &fpDx[width];
    pidy = &fpDy[width];

    const float *const rowEnd = &fpSrc[last];

    for (rowy=&fpSrc[width];
         rowy<rowEnd;
         rowy+=width) {
      *pidx=(rowy[1] - rowy[0]);
      *pidy=(rowy[width] - rowy[-width]);

      ++pidx;
      ++pidy;

      const float *const colEnd = &rowy[w2];
      for (colx=rowy;
           colx<colEnd;
           ++colx,++pidx,++pidy) {
        *pidx=(colx[2] - *colx);
        *pidy=(colx[nextRow] - colx[lastRow]);
      }

      *pidx=(colx[1] - *colx);
      *pidy=(colx[nextRow] - colx[lastRow]);

      ++pidx;
      ++pidy;
    }

    // bottom-left corner
    fpDx[last]=(fpSrc[last+1]-fpSrc[last]);
    fpDy[last]=(fpSrc[last]-fpSrc[last-width]);

    // bottom
    pidx = &fpDx[last+1];
    pidy = &fpDy[last+1];

    const float *const colEnd = &rowEnd[w2];
    for (colx=rowEnd;
         colx<colEnd;
         ++colx,++pidy,++pidx) {
      *pidx=(colx[2]-*colx);
      *pidy=(colx[1]-colx[lastRow]);
    }

    // bottom-right corner
    fpDx[last+w1]=(fpSrc[last+w1]-fpSrc[last+w2]);
    fpDy[last+w1]=(fpSrc[last+w1]-fpSrc[last-width+w1]);

    return true;
  };


  // IntePixel gradient calculation
  bool
  colorContrastGradient::xyDifferentiateImageCartInterPixel(const channel& src,
                                                 channel& dx,
                                                 channel& dy) const {

    const int iWidth  = src.columns();
    const int iHeight = src.rows();
    
    int iX,iY;

    // main block
    for (iY=0; iY< src.lastRow(); iY++) {
      for (iX=0; iX< src.lastColumn(); iX++) {
        dx.at(iY,iX) = src.at(iY+1,iX+1) - src.at(iY,iX);
        dy.at(iY,iX) = src.at(iY+1,iX)   - src.at(iY,iX+1);
      }
    }
    
    // last column
    for (iY=0; iY < src.lastRow(); iY++){
      iX=iWidth-1;
      dy.at(iY,iX)=dx.at(iY,iX) = src.at(iY+1,iX)-src.at(iY,iX);
    }
    
    // last row
    for (iX=0; iX< iWidth-1; iX++){
      iY=iHeight-1;
      dx.at(iY,iX)=src.at(iY,iX+1)-src.at(iY,iX);
      dy.at(iY,iX)=src.at(iY,iX)  -src.at(iY,iX+1);
    }

    return true;
  }


}

