/*
 * Copyright (C) 1998, 1999, 2000, 2001
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 *
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.
 */


/*----------------------------------------------------------------
 * project ....: LTI Digital Image/Signal Processing Library
 * file .......: ltiGradientKernels.h
 * authors ....: Pablo Alvarado
 * organization: LTI, RWTH Aachen
 * creation ...: 11.07.00
 * revisions ..: $Id: ltiGradientKernels.h,v 1.6 2003/11/24 06:53:05 alvarado Exp $
 */

#ifndef _LTI_GRADIENTKERNELS_H_
#define _LTI_GRADIENTKERNELS_H_

#include "ltiLinearKernels.h"

namespace lti {
  /**
   * Two-dimensional separable filter kernels for the gradient approximation
   *
   * This class will create a two-dimensional gradient filter kernel.
   *
   * There are three posibilities for the kernel dimensions 3x3, 4x4 or 5x5.
   * The approximation is done using the suggestions of Ando in its paper
   * "Consistent Gradient Operators" (IEEE PAMI, Vol. 22, No. 3, March 2000)
   *
   * These suggested kernels are separable and are of course implemented that
   * way (much faster!)
   *
   * This object will correspond to the derivative respect the x-Axis.
   * For the y-Axis just exchange the 1D filter kernels, or use the
   * lti::gradientKernelY.
   *
   * The template parameter T should be the same as the value_type of
   * the matrix or channel to be filtered (channel::value_type, double, etc.)
   *
   * It is recomended to use this kernels with floating point matrices
   * or channels, due to precision aspects.
   *
   * @ingroup gLinearFilters
   */
  template<class T>
  class gradientKernelX : public sepKernel<T> {
  public:
    /**
     * constructor
     * @param size is the dimension of the one dimensional part (i.e. the
     *             filter kern is a size x size kernel!)
     *             The valid values are 3, 4 or 5.  If an invalid value is
     *             given, an empty kernel will be created.
     */
    gradientKernelX(const int& size = 3);

    /**
     * initialize this kernel with the specified values
     * @param size size of the kernel in one dimension
     */
    void generate(const int& size);
  };

  /**
   * Two-dimensional separable filter kernels for the gradient approximation
   *
   * This class will create a two-dimensional gradientian filter kernel.
   *
   * There are three posibilities for the kernel dimensions 3x3, 4x4 or 5x5.
   * The approximation is done using the suggestions of Ando in its paper
   * "Consistent Gradient Operators" (IEEE PAMI, Vol. 22, No. 3, March 2000)
   *
   * These suggested kernels are separable and are of course implemented that
   * way (much faster!)
   *
   * This object will correspond to the derivative respect the y-Axis.
   * For the x-Axis just exchange the 1D filter kernels, or use the
   * lti::gradientKernelX .
   *
   * The template parameter T should be the same as the value_type of
   * the matrix or channel to be filtered (channel::value_type, double, etc.)
   *
   * It is recomended to use this kernels with floating point matrices
   * or channels, due to precision aspects.
   *
   * @ingroup gLinearFilters
   */
  template<class T>
  class gradientKernelY : public sepKernel<T> {
  public:
    /**
     * constructor
     *
     * @param size is the dimension of the one dimensional part
     *             (i.e. the filter kern is a size x size kernel!)  The valid
     *             values are 3, 4 or 5.  If an invalid value is given, an
     *             empty kernel will be created.
     */
    gradientKernelY(const int& size = 3);

    /**
     * initialize this kernel with the specified values
     * @param size size of the kernel in one dimension.
     */
    void generate(const int& size);
  };

}

#include "ltiGradientKernels_template.h"

#endif

