/*
 * Copyright (C) 1998, 1999, 2000, 2001
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 *
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.
 */


/* -------------------------------------------------------------------
 * project ....: LTI-Lib: Image Processing and Computer Vision Library
 * file .......: ltiLocalColorFeature.cpp
 * authors ....: Axel Berner, Pablo Alvarado
 * organization: LTI, RWTH Aachen
 * creation ...: 17.1.2002
 * revisions ..: $Id: ltiLocalColorFeature.cpp,v 1.9 2003/10/30 10:02:31 alvarado Exp $
 */

#include "ltiLocalColorFeature.h"

//#define _LTI_DEBUG

namespace lti {
  // --------------------------------------------------
  // localColorFeature::parameters
  // --------------------------------------------------

  // default constructor
  localColorFeature::parameters::parameters()
    : featureExtractor::parameters() {
    slices = 6;
    overlap = true;
    fine = false;
    anglePrecision = 1.0;
  }

  // copy constructor
  localColorFeature::parameters::parameters(const parameters& other)
    : featureExtractor::parameters()  {
    copy(other);
  }

  // destructor
  localColorFeature::parameters::~parameters() {
  }

  // get type name
  const char* localColorFeature::parameters::getTypeName() const {
    return "localColorFeature::parameters";
  }

  // copy member

  localColorFeature::parameters&
    localColorFeature::parameters::copy(const parameters& other) {
# ifndef _LTI_MSC_VER
    // MS Visual C++ 6 is not able to compile this...
    featureExtractor::parameters::copy(other);
# else
    // ...so we have to use this workaround.
    // Conditional on that, copy may not be virtual.
    featureExtractor::parameters& (featureExtractor::parameters::* p_copy)
      (const featureExtractor::parameters&) =
      featureExtractor::parameters::copy;
    (this->*p_copy)(other);
# endif

    slices = other.slices;
    overlap = other.overlap;
    fine = other.fine;
    anglePrecision = other.anglePrecision;

    return *this;
  }

  // alias for copy member
  localColorFeature::parameters&
    localColorFeature::parameters::operator=(const parameters& other) {
    return copy(other);
  }

  // clone member
  functor::parameters* localColorFeature::parameters::clone() const {
    return new parameters(*this);
  }

  /*
   * write the parameters in the given ioHandler
   * @param handler the ioHandler to be used
   * @param complete if true (the default) the enclosing begin/end will
   *        be also written, otherwise only the data block will be written.
   * @return true if write was succeful
   */
# ifndef _LTI_MSC_VER
  bool localColorFeature::parameters::write(ioHandler& handler,
                                         const bool& complete) const
# else
  bool localColorFeature::parameters::writeMS(ioHandler& handler,
                                           const bool& complete) const
# endif
  {
    bool b = true;
    if (complete) {
      b = handler.writeBegin();
    }

    if (b) {
      lti::write(handler,"slices",slices);
      lti::write(handler,"overlap",overlap);
      lti::write(handler,"fine",fine);
      lti::write(handler,"anglePrecision",anglePrecision);

    }

# ifndef _LTI_MSC_VER
    // This is the standard C++ code, which MS Visual C++ 6 is not able to
    // compile...
    b = b && featureExtractor::parameters::write(handler,false);
# else
    bool (featureExtractor::parameters::* p_writeMS)(ioHandler&,const bool&) const =
      featureExtractor::parameters::writeMS;
    b = b && (this->*p_writeMS)(handler,false);
# endif

    if (complete) {
      b = b && handler.writeEnd();
    }

    return b;
  }

# ifdef _LTI_MSC_VER
  bool localColorFeature::parameters::write(ioHandler& handler,
                                         const bool& complete) const {
    // ...we need this workaround to cope with another really awful MSVC bug.
    return writeMS(handler,complete);
  }
# endif

  /*
   * read the parameters from the given ioHandler
   * @param handler the ioHandler to be used
   * @param complete if true (the default) the enclosing begin/end will
   *        be also read, otherwise only the data block will be read.
   * @return true if write was succeful
   */
# ifndef _LTI_MSC_VER
  bool localColorFeature::parameters::read(ioHandler& handler,
                                        const bool& complete)
# else
  bool localColorFeature::parameters::readMS(ioHandler& handler,
                                          const bool& complete)
# endif
  {
    bool b = true;
    if (complete) {
      b = handler.readBegin();
    }

    if (b) {
      lti::read(handler,"slices",slices);
      lti::read(handler,"overlap",overlap);
      lti::read(handler,"fine",fine);
      lti::read(handler,"anglePrecision",anglePrecision);
    }

# ifndef _LTI_MSC_VER
    // This is the standard C++ code, which MS Visual C++ 6 is not able to
    // compile...
    b = b && featureExtractor::parameters::read(handler,false);
# else
    bool (featureExtractor::parameters::* p_readMS)(ioHandler&,const bool&) =
      featureExtractor::parameters::readMS;
    b = b && (this->*p_readMS)(handler,false);
# endif

    if (complete) {
      b = b && handler.readEnd();
    }

    return b;
  }

# ifdef _LTI_MSC_VER
  bool localColorFeature::parameters::read(ioHandler& handler,
                                        const bool& complete) {
    // ...we need this workaround to cope with another really awful MSVC
    // bug.
    return readMS(handler,complete);
  }
# endif

  // --------------------------------------------------
  // localColorFeature
  // --------------------------------------------------

  // default constructor
  localColorFeature::localColorFeature()
    : featureExtractor(){

    // create an instance of the parameters with the default values
    parameters defaultParameters;
    // set the default parameters
    setParameters(defaultParameters);

    cacheVectorSize = 0;

  }

  // copy constructor
  localColorFeature::localColorFeature(const localColorFeature& other)
    : featureExtractor()  {
    copy(other);
  }

  // destructor
  localColorFeature::~localColorFeature() {
    clearCache();
  }

  // returns the name of this type
  const char* localColorFeature::getTypeName() const {
    return "localColorFeature";
  }

  // which apply is faster
  functor::eFasterApply localColorFeature::theFaster() const {
    return OnCopy;
  }

  // copy member
  localColorFeature&
  localColorFeature::copy(const localColorFeature& other) {
    featureExtractor::copy(other);

    cacheVectorSize = other.cacheVectorSize;
    clearCache();

    return (*this);
  }

  // alias for copy member
  localColorFeature&
    localColorFeature::operator=(const localColorFeature& other) {
    return (copy(other));
  }


  // clone member
  functor* localColorFeature::clone() const {
    return new localColorFeature(*this);
  }

  // return parameters
  const localColorFeature::parameters&
    localColorFeature::getParameters() const {
    const parameters* par =
      dynamic_cast<const parameters*>(&functor::getParameters());
    if(isNull(par)) {
      throw invalidParametersException(getTypeName());
    }
    return *par;
  }

  float kompl(const float& f) {
    return 1-f;
  }

  bool localColorFeature::getMasks(const double& angle,
                                   const double& radius,
                                   std::vector<lti::channel8>& rawMasks,
                                   std::vector<lti::channel>& fineMasks) {
    const unsigned int slices = getParameters().slices;
    if(slices < 1)
      return false;

    const int lr = static_cast<int>(radius-0.5);
    const int size = 1+2*lr;

    const int N = getParameters().fine ? 8 : 1;
    const int sizeN = N*size;
    const int lrNs = sizeN*sizeN/4;

    std::vector<channel8> tempMasks;
    tempMasks.resize(slices);
    rawMasks .resize(slices); // full circle + slices
    fineMasks.resize(slices); // full circle + slices
    unsigned int i;

    // create tempMasks (a N-times bigger mask for aliasing)
    for(i=0;i<tempMasks.size();i++) // init and clear
      tempMasks[i].resize(sizeN,sizeN,0,0);
    static const double pi2 = 2*Pi;
    const float start = -(sizeN/2 - 0.5f);
    const float end   =  (sizeN/2 - 0.5f);
    float fx,fy;
    for(fx=start;fx<=end;fx++)
      for(fy=start;fy<=end;fy++)
        if(fx*fx + fy*fy < lrNs){
          // 0.5/radius: to force that every slice has a const number
          // of pixels after rounding
          double a = atan2(fy,fx) - angle + 0.5/radius;
          // force a in [0,2Pi]
          if(a<0)
            a += pi2;
          else if (a>pi2)
            a -= pi2;
          // determ to which slice the pixel belongs to
          const int slice = int(a/pi2*slices);
          tempMasks[slice].at(int(fy+end),int(fx+end)) = 255;
        }

    if (!getParameters().fine) { // speed up if fineMasks aren't needed
      rawMasks = tempMasks;
      return true;
    }

    //create fineMasks
    for(i=0;i<slices;i++)
      fineMasks[i].resize(size,size,0,0.0);
    int x,y,m,n;
    for(i=0;i<slices;i++)
      for(x=0;x<size;x++)
        for(y=0;y<size;y++) {
          for(m=0;m<N;m++)
            for(n=0;n<N;n++){
              if(tempMasks[i].at(y*N+n,x*N+m) != 0)
                fineMasks[i].at(y,x)++;
            }
          fineMasks[i].at(y,x)/=(N*N);
        }

    // create rawMasks
    for(i=0;i<slices;i++)
      rawMasks[i].resize(size,size,0,0);
    lti::dvector v(slices+1);
    for(x=0;x<size;x++)
      for(y=0;y<size;y++) {
        // slice with the highest value is the winner of the pixel
        for(i=0;i<=slices;i++)
          v.at(i) = fineMasks[i].at(y,x);
        const int maxI = v.getIndexOfMaximum();
        if(maxI != 0) {
          rawMasks[maxI].at(y,x) = 1;
        }
      }

    //fineMasks := diffMasks (fine-raw)
    for(i=0;i<slices;i++)
      for(x=0;x<size;x++)
        for(y=0;y<size;y++)
          fineMasks[i].at(y,x) -= rawMasks[i].at(y,x);

    return true;
  }


  bool localColorFeature::maskImage(const image& img,
                                    const lti::channel8& rawMask,
                                    const lti::channel& fineMask,
                                    const tpoint<float>& position,
                                    double& red,
                                    double& green,
                                    double& blue) {
    tpoint<int> pos;
    pos.castFrom(position);
    const int startX = max(0           ,pos.x-rawMask.columns()/2);
    const int endX =   min(img.size().x,pos.x+rawMask.columns()/2+1);
    const int startY = max(0           ,pos.y-rawMask.rows()   /2);
    const int endY =   min(img.size().y,pos.y+rawMask.rows()    /2+1);
    tpoint<int> startP(pos.x-rawMask.columns()/2,pos.y-rawMask.rows()/2);
    float fac,counter;
    red = green = blue = counter = 0.0;
    tpoint<int> p;
    for(p.x=startX;p.x<endX;p.x++)
      for(p.y=startY;p.y<endY;p.y++)
        if(rawMask.at(p-startP) != 0) {
          red   += img.at(p).getRed();
          green += img.at(p).getGreen();
          blue  += img.at(p).getBlue();
          counter++;
        }
    if (getParameters().fine)
      for(p.x=startX;p.x<endX;p.x++)
        for(p.y=startY;p.y<endY;p.y++) {
          fac = fineMask.at(p-startP);
          if(0.0 != fac) {
            red   += fac * img.at(p).getRed();
            green += fac * img.at(p).getGreen();
            blue  += fac * img.at(p).getBlue();
            counter += fac;
          }
        }
    red /=counter;
    green /= counter;
    blue /= counter;
    return true;
  }


  // -------------------------------------------------------------------
  // slice mask holder class
  // -------------------------------------------------------------------
  localColorFeature::sliceMaskHolder::sliceMaskHolder()
    : mask(),pos(0),in(),out() {
  }

  localColorFeature::sliceMaskHolder::sliceMaskHolder(const double radius,
                                                      const double angle,
                                                      const int slices) {
    generate(radius,angle,slices);
  }

  localColorFeature::sliceMaskHolder::sliceMaskHolder(const
                                                      sliceMaskHolder&
                                                      other)
    : mask()  {
    copy(other);
  }

  localColorFeature::sliceMaskHolder&
  localColorFeature::sliceMaskHolder::copy(const sliceMaskHolder& other) {
    mask.copy(other.mask);
    pos = other.pos;
    in.copy(other.in);
    out.copy(other.out);
    return *this;
  }

  localColorFeature::sliceMaskHolder&
  localColorFeature::sliceMaskHolder::operator=(const sliceMaskHolder&
                                                other) {
    return copy(other);
  }

  void localColorFeature::sliceMaskHolder::generate(const double radius,
                                                    const double angle,
                                                    const int slices) {

    // assuming that a radius 1 means just one pixel...
    int size = static_cast<int>(ceil(radius));
    pos = 1-size; // the position of the first point of the mask
    size -= pos;

    // the mask will be size x size
    mask.resize(size,size,255,false,true);

    // the in/out points must also have "size" elements
    in.resize(size,size,false,true);
    out.resize(size,-1,false,true);

    // create the io points and the mask
    int i,j,idx,y;
    const int r2 = pos*pos;
    double pa;
    double a = angle;
    int o;
    while (a > (2*Pi)) {
      a -= 2*Pi;
    }
    while (a < 0) {
      a += 2*Pi;
    }

    static const double pi2 = 2*Pi;

    // go through the y-axis
    for (i=0;i<in.size();++i) {
      // compute y^2
      y = i+pos;
      y *= y;

      // the in and out points correspond to the borders of
      // a circle:
      in.at(i)  = -iround(sqrt(r2-y))-pos;
      out.at(i) = o = iround(sqrt(r2-y))-pos;

      // for all points between the in and out points
      for (j=in.at(i);j<=o;++j) {
        // compute the angle of this point
        pa = atan2(double(i+pos),double(j+pos))-a;

        // fix the angle to be between 0 and 2*Pi
        while (pa < 0) {
          pa += pi2;
        }
        while (pa > pi2) {
          pa -= pi2;
        }

        // compute the index of the splice
        idx = static_cast<int>(pa*slices/pi2);
        // and indicate the computation in the mask.
        mask.at(i,j)=idx;
      }
    }
  }

  // -------------------------------------------------------------------
  // The apply-methods!
  // -------------------------------------------------------------------

  // On copy apply for type std::list<location>!
  bool localColorFeature::apply(const lti::image& src,
                                const std::list<location>& locs,
                                std::list<dvector>& dest) {
    const parameters& param = getParameters();
    if (param.fine) {
      return getFineFeatures(src,locs,dest);
    } else {
      return getRawFeatures(src,locs,dest);
    }
  }

  // On copy apply for type std::list<location>!
  bool localColorFeature::getFineFeatures(const lti::image& src,
                                          const std::list<location>& locs,
                                          std::list<dvector>& dest) {
    unsigned int i;
    const int slices =  getParameters().slices;
    std::vector<channel8> rawMasks1, rawMasks2;
    std::vector<channel> fineMasks1,fineMasks2;

    std::list<location>::const_iterator locList;
    std::list<dvector>::iterator destList;
    dest.resize(locs.size());

    // for all locations
    for(locList = locs.begin(),destList = dest.begin();
        locList!=locs.end();
        locList++, destList++){
      // create CutOut-Masks
      getMasks((*locList).angle,(*locList).radius,
               rawMasks1,fineMasks1);
      if (getParameters().overlap) {
        getMasks((*locList).angle+Pi/slices,(*locList).radius,
                 rawMasks2,fineMasks2);
      }

      // calc averagecolors per slice
      // store-structure for e.g. 2 slices:
      // full, slice1, slice2, ... clockwise
      // => rrrrrgggggbbbbb
      if(getParameters().overlap) {
        (*destList).resize((rawMasks1.size()+rawMasks2.size())*3,
                           0.0,false,false);
        const unsigned int rmSize = rawMasks1.size();// = rawMasks2.size();
        for(i=0;i<rmSize;i++) // 0 = full circle
          maskImage(src,rawMasks1[i],fineMasks1[i],
                    (*locList).position,
                    (*destList)[0*(rmSize+rmSize-1)+2*i-1],
                    (*destList)[1*(rmSize+rmSize-1)+2*i-1],
                    (*destList)[2*(rmSize+rmSize-1)+2*i-1]);
        for(i=0;i<rmSize;i++) // 0 = full circle
          maskImage(src,rawMasks2[i],fineMasks2[i],
                    (*locList).position,
                    (*destList)[0*(rmSize+rmSize-1)+2*i],
                    (*destList)[1*(rmSize+rmSize-1)+2*i],
                    (*destList)[2*(rmSize+rmSize-1)+2*i]);

      } else { // not overlapping
        (*destList).resize(rawMasks1.size()*3,0.0,false,false);
        const unsigned int rmSize = rawMasks1.size();
        for(i=0;i<rmSize;i++) // 0 = full circle
          maskImage(src,rawMasks1[i],fineMasks1[i],
                    (*locList).position,
                    (*destList)[0*rmSize+i],
                    (*destList)[1*rmSize+i],
                    (*destList)[2*rmSize+i]);
      }
    } //end for all locations
    return true;
  };

  /*
   * generate the feature %vector given an %image, a
   * sliceMaskHolder %object and the position of a location,
   * generate the feature vector.
   */
  bool localColorFeature::maskImage(const image& img,
                                    const sliceMaskHolder& msk,
                                    const tpoint<float>& pos,
                                    dvector& red,
                                    dvector& green,
                                    dvector& blue) {

    // check if the mask lies on the borders of the image
    const int border = -msk.pos;
    const int slices = getParameters().slices;
    const point p(iround(pos.x),iround(pos.y));
    const point bpmi(p.x-border,p.y-border);
    const point bpma(p.x+border,p.y+border);
    int x,y,lasty,lastx,i,j,idx;
    rgbPixel px;

    dvector n(slices,0.0);
    red.resize(slices,0.0,false,true);
    green.resize(slices,0.0,false,true);
    blue.resize(slices,0.0,false,true);

    if ((p.x >= border) && (p.x < img.columns()-border) &&
        (p.y >= border) && (p.y < img.rows()-border)) {
      // no problem!  location is inside the image

      for (i=0,y=bpmi.y,lasty=bpma.y;y<=lasty;++y,++i) {
        for (j=msk.in.at(i),x=j+bpmi.x,lastx=msk.out.at(i);
             j<=lastx;++x,++j) {
          px = img.at(y,x);
          idx = msk.mask.at(i,j);
          n.at(idx)++;
          red.at(idx)   += double(px.getRed());
          green.at(idx) += double(px.getGreen());
          blue.at(idx)  += double(px.getBlue());
        }
      }
      red.edivide(n);
      green.edivide(n);
      blue.edivide(n);
    } else {
      // upps... we better check the boundaries of the region
      for (i=0,y=bpmi.y,lasty=bpma.y;y<=lasty;++y,++i) {
        for (j=msk.in.at(i),x=j+bpmi.x,lastx=msk.out.at(i);
             j<=lastx;++x,++j) {
          if ((x>=0) && (x<img.columns()) &&
              (y>=0) && (y<img.rows())) {
            px = img.at(y,x);
            idx = msk.mask.at(i,j);
            n.at(idx)++;
            red.at(idx)   += double(px.getRed());
            green.at(idx) += double(px.getGreen());
            blue.at(idx)  += double(px.getBlue());
          }
        }
      }
      red.edivide(n);
      green.edivide(n);
      blue.edivide(n);
    }

    return true;
  }

  void localColorFeature::clearCache() {
    std::map<float, std::vector< sliceMaskHolder* > >::iterator cit;
    std::vector< sliceMaskHolder* >::iterator vit;

    for (cit=cache.begin();cit!=cache.end();++cit) {
      for (vit = (*cit).second.begin();
           vit != (*cit).second.end();
           ++vit) {
        delete (*vit);
        (*vit) = 0;
      }
    }

    cache.clear();
  }

  // On copy apply for type std::list<location>!
  bool localColorFeature::getRawFeatures(const lti::image& src,
                                         const std::list<location>& locs,
                                         std::list<dvector>& dest) {

    const parameters& param = getParameters();

    std::list<location>::const_iterator lit;
    std::list<dvector>::iterator dit;
    dest.resize(locs.size());

    // a cache for radius and angle
    // the map will be addressed with the radius of the location
    // and each vector will
    std::map<float, std::vector< sliceMaskHolder* > >::iterator cit;

    // number of angles per radius depends on the user specified
    // precision
    const int numAngles(static_cast<int>(0.5 + 360.0/param.anglePrecision));

    // if cache contains the wrong vectors, clear it and begin new
    if (numAngles != cacheVectorSize) {
      clearCache();
      cacheVectorSize = numAngles;
    }

    // constant used for the angle conversion
    static const float pi2 = static_cast<float>(2.0*Pi);
    const float acc = float(numAngles)/pi2;
    int aidx,aidxo,i,i2;
    sliceMaskHolder* sliceHolder = 0;
    sliceMaskHolder* sliceHolderOverlap = 0;

    dvector red,green,blue;

    float ang = 0;

    // for each location
    for (lit = locs.begin(),dit = dest.begin();
         lit != locs.end(); ++lit,++dit) {

      // get the corresponding sliceMaskHolder
      cit = cache.find((*lit).radius);

      if (cit == cache.end()) {
        // there is no entry for the given radius yet... create one
        cache[(*lit).radius] =
          std::vector<sliceMaskHolder*>(numAngles,(sliceMaskHolder*)0);
        cit = cache.find((*lit).radius);
      }

      // which angle entry should be checked?
      ang = (*lit).angle;
      while (ang<0) {
        ang+=pi2;
      }
      while (ang>pi2) {
        ang-=pi2;
      }

      aidx = static_cast<int>(ang*acc);

      // check if a mask for this angle and radius was already computed
      sliceHolder = (*cit).second[aidx];
      if (sliceHolder == 0) {
        // there is no mask for this angle and radius yet.
        // Compute a new one
        sliceHolder = new sliceMaskHolder((*lit).radius,
                               aidx*pi2/numAngles,
                               param.slices);
        (*cit).second[aidx] = sliceHolder;
      }

      // generate the feature vector
      maskImage(src,*sliceHolder,(*lit).position,red,green,blue);

      if (param.overlap) {
        // which angle entry should be checked?
        ang = (*lit).angle+static_cast<float>(Pi/param.slices);
        while (ang<0) {
          ang+=pi2;
        }
        while (ang>pi2) {
          ang-=pi2;
        }

        aidxo = static_cast<int>(ang*acc);

        // a second slice holder is required with a new angle
        sliceHolderOverlap = (*cit).second[aidxo];
        if (sliceHolderOverlap == 0) {
          // there is no mask for this angle and radius yet.
          // Compute a new one
          sliceHolderOverlap = new sliceMaskHolder((*lit).radius,
                                                   aidxo*pi2/numAngles,
                                                   param.slices);
          (*cit).second[aidx] = sliceHolderOverlap;
        }

        const int goff = 2*red.size();
        const int boff = 4*red.size();
        dvector ro,go,bo;

        // generate the "overlapping" feature vector
        maskImage(src,*sliceHolderOverlap,(*lit).position,ro,go,bo);

        // and concatenate all values
        (*dit).resize(6*red.size(),0.0,false,false);
        for (i=0,i2=0;i<red.size();++i,i2+=2) {
          (*dit).at(i2) =  red.at(i);
          (*dit).at(i2+1) = ro.at(i);

          (*dit).at(i2+goff) = green.at(i);
          (*dit).at(i2+goff+1) =  go.at(i);

          (*dit).at(i2+boff) = blue.at(i);
          (*dit).at(i2+boff+1) = bo.at(i);
        }
      } else {
        // no overlap -> just concatenate the vectors
        (*dit).resize(3*red.size(),0.0,false,false);
        (*dit).fill(red);
        (*dit).fill(green,red.size());
        (*dit).fill(blue,red.size()*2);
      }

    }

    return true;
  }

}





