/*
 * Copyright (C) 1998, 1999, 2000, 2001
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 *
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.
 */


/*--------------------------------------------------------------------
 * project ....: LTI-Lib: Image Processing and Computer Vision Library
 * file .......: ltiLocalColorFeature.h
 * authors ....: Axel Berner
 * organization: LTI, RWTH Aachen
 * creation ...: 17.1.2002
 * revisions ..: $Id: ltiLocalColorFeature.h,v 1.5 2003/10/30 10:02:31 alvarado Exp $
 */

#ifndef _LTI_LOCAL_COLOR_FEATURE_H_
#define _LTI_LOCAL_COLOR_FEATURE_H_

#include "ltiObject.h"

#include <list>
#include <map>
#include <vector>

#include "ltiImage.h"
#include "ltiLocation.h"
#include "ltiFeatureExtractor.h"

namespace lti {
  /**
   * This functor calculates for each location in a given list
   * color-features.
   *
   * Each location will be considere as a circular region, that will
   * be divide in a specified number of splices.  The color feature will
   * compute for each splice the average color.
   *
   * It is also possible to compute overlapping splices, to make the feature
   * more robust against noise.
   */
  class localColorFeature : public featureExtractor {
  public:
    /**
     * the parameters for the class localColorFeature
     */
    class parameters : public featureExtractor::parameters {
    public:
      /**
       * default constructor
       */
      parameters();

      /**
       * copy constructor
       * @param other the parameters object to be copied
       */
      parameters(const parameters& other);

      /**
       * destructor
       */
      ~parameters();

      /**
       * returns name of this type
       */
      const char* getTypeName() const;

      /**
       * copy the contents of a parameters object
       * @param other the parameters object to be copied
       * @return a reference to this parameters object
       */
      parameters& copy(const parameters& other);

      /**
       * copy the contents of a parameters object
       * @param other the parameters object to be copied
       * @return a reference to this parameters object
       */
      parameters& operator=(const parameters& other);


      /**
       * returns a pointer to a clone of the parameters
       */
      virtual functor::parameters* clone() const;

      /**
       * write the parameters in the given ioHandler
       * @param handler the ioHandler to be used
       * @param complete if true (the default) the enclosing begin/end will
       *        be also written, otherwise only the data block will be written.
       * @return true if write was successful
       */
      virtual bool write(ioHandler& handler,const bool& complete=true) const;

      /**
       * read the parameters from the given ioHandler
       * @param handler the ioHandler to be used
       * @param complete if true (the default) the enclosing begin/end will
       *        be also written, otherwise only the data block will be written.
       * @return true if write was successful
       */
      virtual bool read(ioHandler& handler,const bool& complete=true);

#     ifdef _LTI_MSC_VER
      /**
       * this function is required by MSVC only, as a workaround for a
       * very awful bug, which exists since MSVC V.4.0, and still by
       * V.6.0 with all bugfixes (so called "service packs") remains
       * there...  This method is also public due to another bug, so please
       * NEVER EVER call this method directly: use read() instead
       */
      bool readMS(ioHandler& handler,const bool& complete=true);

      /**
       * this function is required by MSVC only, as a workaround for a
       * very awful bug, which exists since MSVC V.4.0, and still by
       * V.6.0 with all bugfixes (so called "service packs") remains
       * there...  This method is also public due to another bug, so please
       * NEVER EVER call this method directly: use write() instead
       */
      bool writeMS(ioHandler& handler,const bool& complete=true) const;
#     endif

      // ------------------------------------------------
      // the parameters
      // ------------------------------------------------

      /**
       * A given location will be interpreted as a circle, that will
       * be divided in this number of slices.  The minimal allowed
       * value of slices is one, and the maximal allowed slices value
       * is 120.
       *
       * The number of dimensions of the output feature vector will be
       * "slices" or "2*slices" depending on the "overlap" parameter.
       *
       * Default value: 6
       */
      unsigned int slices;

      /**
       * overlapping slices.  If true, twice the number of slices will be
       * computed and the feature vector will have 2*slices elements.  If
       * false, single slices will be computed and the number of elements
       * of the feature vector will be just 1*slices.
       *
       * Default value: true
       */
      bool overlap;

      /**
       * use aliased slice-borders.  If activated (true), the computation time
       * will increase considerably (very slow!)
       *
       * Default value: false
       */
      bool fine;

      /**
       * If this value is greater 0, only angles that are a multiple of
       * this anglePrecision value will be allowed.  Other values will be
       * rounded to valid values.  If zero or negative, all angles are
       * possible.
       *
       * The value must be given in degrees (and NOT in radians!).
       *
       * Default value: 1 degree (= Pi/180 radians)
       */
      double anglePrecision;
    };

    /**
     * default constructor
     */
    localColorFeature();

    /**
     * copy constructor
     * @param other the object to be copied
     */
    localColorFeature(const localColorFeature& other);

    /**
     * destructor
     */
    virtual ~localColorFeature();

    /**
     * returns the name of this type ("localColorFeature")
     */
    virtual const char* getTypeName() const;


    /**
     * @param src original image to be analyzed.
     * @param locs locations that need to be analyzed.
     * @param dest list of feature vectors.  This list will have the
     *             same size as locs and contains the average color of the
     *             location and of the slices.   The structure of an element
     *             of the vector list for slices=2 and overlaping would be:
     *             rrrrggggbbbb.
     * @return true if successful, or false otherwise.  */
    bool apply(const lti::image& src,
               const std::list<location>& locs,
               std::list<dvector>& dest);

    /**
     * returns whether the onPlace version of apply is faster or
     * the onCopy version.
     */
    virtual functor::eFasterApply theFaster() const;

    /**
     * copy data of "other" functor.
     * @param other the functor to be copied
     * @return a reference to this functor object
     */
    localColorFeature& copy(const localColorFeature& other);

    /**
     * alias for copy member
     * @param other the functor to be copied
     * @return a reference to this functor object
     */
    localColorFeature& operator=(const localColorFeature& other);

    /**
     * returns a pointer to a clone of this functor.
     */
    virtual functor* clone() const;

    /**
     * @return used parameters
     */
    const parameters& getParameters() const;

    /**
     * clear the internal cache used to accelerate the computation of
     * the slice masks.
     */
    void clearCache();

  protected:

    /**
     * simple structure used to hold a mask and the correspoinding io-points
     * of the circle.
     */
    class sliceMaskHolder {
    public:
      /**
       * default constructor
       */
      sliceMaskHolder();

      /**
       * constructor to generate directly the mask
       */
      sliceMaskHolder(const double radius,
                      const double angle,
                      const int slices);

      /**
       * copy constructor
       */
      sliceMaskHolder(const sliceMaskHolder& other);

      /**
       * generate this mask for the given %parameters
       */
      void generate(const double radius,
                    const double angle,
                    const int slices);

      /**
       * copy member
       */
      sliceMaskHolder& copy(const sliceMaskHolder& other);

      /**
       * alias for copy member
       */
      sliceMaskHolder& operator=(const sliceMaskHolder& other);

      /**
       * mask contains the index of the slice for a pixel
       */
      channel8 mask;

      /**
       * pos contains the position of the first element in the io points
       */
      int pos;

      /**
       * input points.  The size of this vector will always be the number
       * of rows of the mask.
       */
      ivector in;

      /**
       * output points.  The size of this vector will always be the number
       * of rows of the mask.
       */
      ivector out;
    };

    /**
     * creates masks (circle, slices) with the @param angle and @param radius
     * @param angle orieantation of the location
     * @param radius radius of the location
     * @param rawMasks binary masks
     * @param fineMasks different-masks from the rawMasks to the aliased-masks
     * @return true if successful, of false otherwise.
     */
    bool getMasks(const double& angle,
                  const double& radius,
                  std::vector<channel8>& rawMasks,
                  std::vector<channel>& fineMasks);

    /**
     * calculates the average color of the image img at the given
     * position with the mask
     * @param img the image, from which averages will be computed.
     * @param rawMask the raw mask
     * @param fineMask the fine mask
     * @param position in the image, where the average will be computed.
     * @param red component
     * @param green component
     * @param blue component
     * @return true if successful, of false otherwise.
     */
    bool maskImage(const image& img,
                   const lti::channel8& rawMask,
                   const lti::channel& fineMask,
                   const tpoint<float>& position,
                   double& red ,double& green, double& blue);

    /**
     * generate the feature %vector given an %image, a
     * sliceMaskHolder %object and the position of a location,
     * generate the feature vector.
     */
    bool maskImage(const image& img,
                   const sliceMaskHolder& msk,
                   const tpoint<float>& pos,
                   dvector& red,
                   dvector& green,
                   dvector& blue);

    /**
     * @param src original image to be analyzed.
     * @param locs locations that need to be analyzed.
     * @param dest list of feature vectors.  This list will have the same
     * size as locs and contains the average color of the location and of
     * the slices.
     * structur of an element of the vector list (e.g. 2 slices):
     * rrrrrgggggbbbbb.
     * @return true if successful, or false otherwise.
     */
    bool getFineFeatures(const lti::image& src,
                         const std::list<location>& locs,
                         std::list<dvector>& dest);

    /**
     * @param src original image to be analyzed.
     * @param locs locations that need to be analyzed.
     * @param dest list of feature vectors.  This list will have the same
     * size as locs and contains the average color of the location and of
     * the slices.
     * structur of an element of the vector list (e.g. 2 slices):
     * rrrrrgggggbbbbb.
     * @return true if successful, or false otherwise.
     */
    bool getRawFeatures(const lti::image& src,
                        const std::list<location>& locs,
                        std::list<dvector>& dest);

    /**
     * the cache used to store the last computed slice masks
     */
    std::map<float, std::vector< sliceMaskHolder* > > cache;

    /**
     * size of the vectors contained in the cache
     */
    int cacheVectorSize;

  };
}

#endif

