/*
 * Copyright (C) 1998, 1999, 2000, 2001
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 * 
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.  
 */ 

 
/* -------------------------------------------------------------------
 * project ....: LTI-Lib: Image Processing and Computer Vision Library
 * file .......: ltiSplitImageToYUV.cpp
 * authors ....: Volker Schmirgel
 * organization: LTI, RWTH Aachen
 * creation ...: 17.12.2002
 * revisions ..: $Id: ltiSplitImageToYUV.cpp,v 1.1 2003/11/14 12:23:28 goenner Exp $
 */

//TODO: include files
#include "ltiSplitImageToYUV.h"


namespace lti {

  // --------------------------------------------------
  // splitImageToYUV
  // --------------------------------------------------

  // default constructor
  splitImageToYUV::splitImageToYUV()
    : splitImage(){
  }

  // copy constructor
  splitImageToYUV::splitImageToYUV(const splitImageToYUV& other) {
    copy(other);
  }

  // destructor
  splitImageToYUV::~splitImageToYUV() {
  }

  // returns the name of this type
  const char* splitImageToYUV::getTypeName() const {
    return "splitImageToYUV";
  }

  // which apply is faster
  functor::eFasterApply splitImageToYUV::theFaster() const {
    return OnCopy;
  }

  // copy member
  splitImageToYUV&
    splitImageToYUV::copy(const splitImageToYUV& other) {
      splitImage::copy(other);

    return (*this);
  }

  // alias for copy member
  splitImageToYUV&
    splitImageToYUV::operator=(const splitImageToYUV& other) {
    return (copy(other));
  }


  // clone member
  functor* splitImageToYUV::clone() const {
    return (new splitImageToYUV(*this));
  }

  // -------------------------------------------------------------------
  // The apply-methods!
  // -------------------------------------------------------------------

  // split pixel into float values (between 0 and 1)
  void splitImageToYUV::apply(const rgbPixel& pixel,
                              float& c1,
                              float& c2,
                              float& c3) const {

    const float red = float ( pixel.getRed() );
    const float green = float ( pixel.getGreen() );
    const float blue = float ( pixel.getBlue() );

    //   most common values in literature and internet
    //   Y =  0.299*R + 0.587*G + 0.114*B 
    //   U = -0.169*R - 0.331*G + 0.500*B + 128.0 
    //   V =  0.500*R - 0.419*G - 0.081*B + 128.0 
    //   other transformation values can be found but are usually just a result
    //   of nomalization 
    //   Gonzales and Woods only deal with the American YIQ standard

    //   For most accurate results we use values of the  
    //   ITU-RS601 (formerly known as CCIR-601)
    //   value Range here: Y [16...235], U/V [16..240]
    //   Y =  0.256635*R + 0.503971*G + 0.098218*B + 16
    //   U = -0.148135*R - 0.290902*G + 0.439036*B + 128.0 
    //   V =  0.439327*R - 0.367673*G - 0.071655*B + 128.0 
    //   in the norm Cb is used instead of U and Cr instead of V
    //   but here we work with YUV 

    c1 = ( red *0.256635f +
          green *0.503971f +
          blue *0.098218f +16)/255.0f;      // Y
    c2 = (red *(-0.148135f) +
          green *(-0.290902f) +
          blue *0.439036f +128)/255.0f;      // U
    c3 = (red *0.439327f +
          green *(-0.367673f) +
          blue *(-0.071655f)+128)/255.0f;   // V
  }

  // split pixel into 8-bit values (between 0 and 255)
  // N.B.: when casting the transformation result to unsigned shorts
  // (8-bit channel) major rounding errors will occur.
  // As a result, the merging operation might
  // produce negative values or values > 1,  which are truncated subsequently.
  // When accurate Y, U and V values are required, prefer float values!
  void splitImageToYUV::apply(const rgbPixel& pixel,
                              ubyte& c1,
                              ubyte& c2,
                              ubyte& c3) const {

    const float red = float ( pixel.getRed() );
    const float green = float ( pixel.getGreen() );
    const float blue = float ( pixel.getBlue() );

  // see above for explanation of magic numbers
    c1 = (ubyte)(red *0.256635f +
                 green *0.503971f +
                 blue *0.098218f +16);      // Y 
    c2 = (ubyte)(red *(-0.148135f) +
                 green *(-0.290902f) +
                 blue *0.439036f +128);      // U
    c3 = (ubyte)(red *0.439327f +
                 green *(-0.367673f) +
                 blue *(-0.071655f)+128);   // V

  }



  // split image into float channels
  void splitImageToYUV::apply(const image& img,
                              channel& c1,
                              channel& c2,
                              channel& c3) const {

    // make the channels size of source image...
    // Todo: don't initialize
    c1.resize(img.rows(),img.columns(),0,false,false);
    c2.resize(img.rows(),img.columns(),0,false,false);
    c3.resize(img.rows(),img.columns(),0,false,false);

    const int numRows (img.rows());
    const int numColumns (img.columns());
    int i;

    for(i=0; i<numRows; i++) {
      const vector<rgbPixel>& srcVct = img.getRow(i);
      vector<float>& c1Vct = c1.getRow(i);
      vector<float>& c2Vct = c2.getRow(i);
      vector<float>& c3Vct = c3.getRow(i);

      int j;
      for(j=0; j<numColumns; j++) {
	// take pixel at position j
	rgbPixel pix ( srcVct.at(j) );  // single Pixel Element in RGB-values...

	apply(pix, c1Vct.at(j), c2Vct.at(j), c3Vct.at(j));

      } // loopcd
    }
  }


  // split image into 8-bit channels
  // N.B.: when casting the transformation result to unsigned shorts
  // (8-bit channel) major rounding errors will occur.
  // As a result, the merging operation might produce negative
  // values or values > 1,  which are truncated subsequently.
  // When accurate Y, U, and V channels are required, prefer float channels!
  void splitImageToYUV::apply(const image& img,
                              channel8& c1,
                              channel8& c2,
                              channel8& c3) const {

    // make the channels size of source image...
    c1.resize(img.rows(),img.columns(),0,false,false);
    c2.resize(img.rows(),img.columns(),0,false,false);
    c3.resize(img.rows(),img.columns(),0,false,false);

    const int numRows (img.rows());
    const int numColumns (img.columns());
    int i;

    for(i=0; i<numRows; i++) {
      const vector<rgbPixel>& srcVct = img.getRow(i);
      vector<ubyte>& c1Vct = c1.getRow(i);
      vector<ubyte>& c2Vct = c2.getRow(i);
      vector<ubyte>& c3Vct = c3.getRow(i);

      int j;
      for(j=0; j<numColumns; j++) {
	// take pixel at position j
	rgbPixel pix ( srcVct.at(j) );  // single Pixel Element in RGB-values...

	apply(pix, c1Vct.at(j), c2Vct.at(j), c3Vct.at(j));

      } // loopcd
    }
  }
}

