/*
 * Copyright (C) 1998, 1999, 2000, 2001
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 *
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.
 */


#include <limits>
#include "ltiEquationSystem.h"
#include "ltiSVD.h"
#include "ltiMath.h"

#undef min

namespace lti {
  //--- matrixInversion ---
  //-----------------------

  /*
  template <class T>
    matrixInversion<T>::parameters::parameters(void)

  template <class T>
  void matrixInversion<T>::parameters::copy(const parameters& other)

  template <class T>
  functor::parameters* matrixInversion<T>::parameters::clone() const
  */

  // use Sqrt((std::numeric_limits<double>::min)()) as epsilon value
  template <class T>
  const T matrixInversion<T>::my_epsilon=sqrt(std::numeric_limits<T>::min());


  template<class T>
  matrixInversion<T>::matrixInversion() {
    parameters tmp;
    setParameters(tmp);
  }


  // onCopy version of apply.
  template<class T>
  bool matrixInversion<T>::apply(const matrix<T>& theMatrix,
                                 matrix<T>& theInverse) const {
    int i;
    const int &n=theMatrix.rows();
    if(n!=theMatrix.columns()) throw incompatibleDimensionsException();
    vector<T> unitVector(n,0.0);

    const parameters& p=getParameters();

    if (p.method == parameters::LUD) {
      luSolution<T> eqs;
      typename luSolution<T>::parameters luSolPar;
      // do not copy (theMatrix could be huge!)
      // const_cast valid since luDecomp and luSol don't change
      // the SystemMatrix
      luSolPar.systemMatrix.useExternData(theMatrix.rows(),
                                          theMatrix.columns(),
                                          const_cast<T*>(&theMatrix.at(0,0)));

      eqs.setParameters(luSolPar);


      theInverse.resize(n,n);
      for(i=0;i<n-1;i++) {
        unitVector[i]=1;
        eqs.apply(unitVector,theInverse[i]);
        unitVector[i]=0;
      }
      unitVector[i]=1;
      eqs.apply(unitVector,theInverse[i]);

      theInverse.transpose();

      return true;
    } else if (p.method == parameters::SVD) {
      singularValueDecomp<T> svd;

      matrix<T> u;
      matrix<T> v;
      vector<T> w;

      if (svd.apply(theMatrix, u,w,v)) {

        theInverse.copy(v);
        for (int i=0; i<theInverse.rows(); i++) {
          T wi=0;
          if (i < w.size()) {
            wi=w.at(i);
          }
          if (abs(wi) < my_epsilon) {
            wi=(wi >= 0) ? T(0) : T(0);
          } else {
            wi=T(1)/wi;
          }
          theInverse.getRow(i).multiply(wi);
        }
        u.transpose();
        theInverse.multiply(u);
        return true;
      } else {
        setStatusString(svd.getStatusString());
        return false;
      }
    } else {
      setStatusString("Unknown inversion method!");
      return false;
    }

  }

  // onPlace version of apply.
  template<class T>
  bool matrixInversion<T>::apply(matrix<T>& theMatrix) const
  {
    matrix<T> tmp;
    bool r=apply(theMatrix,tmp);
    tmp.detach(theMatrix);
    return r;
  }

  template<class T>
  void matrixInversion<T>::useLUD() {
    parameters t;
    t.method=parameters::LUD;
    setParameters(t);
  }

  template<class T>
  void matrixInversion<T>::useSVD() {
    parameters t;
    t.method=parameters::SVD;
    setParameters(t);
  }

  template <class T>
  const typename matrixInversion<T>::parameters&
  matrixInversion<T>::getParameters() const {
    return
      dynamic_cast<const parameters&>(linearAlgebraFunctor::getParameters());
  }
}

